      program as

c----------------------------------------------------------------------
c  as  - estimate the confidence intervals of the proportions of
c        specific agreement (p+ and p-) using the nonparametric
c        bootstrap
c
c        in general the program generates a large number of simulated
c        frequency tables according to user-supplied sampling 
c        probabilities.  it counts in how many of these tables some 
c        criterion function (such as a measure of marginal disparity) 
c        in simulated tables exceeds some reference value (such as the 
c        marginal disparity in observed data).
c
c        the criterion function is calculated in a modifiable
c        subroutine, fcrit
c
c        here the criteria are ps+ and ps-
c
c
c  author:      john uebersax
c  based on:    mh.f
c
c  date:        change(s):
c  ---------    -------------------------------------------------------
c  08 mar o8    adapted from program m2.f
c----------------------------------------------------------------------
c notes:  this version populates simulated tables one obs at a time
c         and should probably only be used for small tables 
c
c         crit = the variable/value against which the simulated
c                tables will be evaluated.  here an ad hoc measure of
c                marginal homogeneity
c
c  This program was compiled and run using Absoft Pro Fortran 90 v. 7
c  as a .f program (not a .f90 program).  This means the program
c  is treated as having a conventional 72 column fortran *format* for
c  command statements, but Fortran 90 *syntax* is permitted.
c----------------------------------------------------------------------

c     ------------------------
c     data declaration section
c     ------------------------

c     run parameters

      integer, parameter ::  nrep = 100000     ! no. of replications
      integer, parameter ::  m    =     2    ! order of table
      integer, parameter ::  ndim = m * m    ! size of table (as vector)
      
c     ===note:  for convenience, the data are used both in vector and table form===

c     for data as a vector  

      real               ::  fo (ndim)       ! observed frequencies
      real               ::  x (ndim)        ! simulated frequencies
      real               ::  p (ndim)        ! cell proportions
      real               ::  c (ndim)        ! cumulative proportions 

      real               ::  vec   ( nrep)   ! results
      real               ::  vec2  ( nrep)   ! sorted results

      integer            ::  icount (ndim)   ! integer version

c     for data as a table

      real               ::  o(m, m)         ! observed frequencies
      real               ::  s(m, m)         ! simulated frequencies
      real               ::  row (m)         ! row marginals
      real               ::  col (m)         ! column marginals

      real               ::  crit            ! criterion value
      real               ::  refcrit         ! reference criterion value
      real               ::  diff (m)        ! differences

      logical            ::  rprint          ! write each simulation result to file?

      rprint = .true.

c     --------------------------------  
c     seed the random number generator
c     --------------------------------

      iseed  = 232
c      iseed = 17

c     save for printing later

      iseed0 = iseed      
      r      = ran0(iseed)

c     ---------------------------  
c     open input and output files
c     ---------------------------

      open (unit=1, file='table.txt'  )
      open (unit=2, file='output.txt' )
      open (unit=3, file='results.txt')
      open (unit=4, file='sorted.txt' )


c     ----------
c     initialize
c     ----------

      fo   = 0.
      x    = 0.
      p    = 0.
      c    = 0.
      o    = 0.
      s    = 0.
      row  = 0.
      col  = 0.
      crit = 0.
      diff = 0.

c     --------------------------------
c     read data table as one-way table
c     --------------------------------

      do i = 1, ndim
         read (1, *) fo(i)
c         write (*,*) fo(i)
      end do

      total = sum(fo(:))
      n     = int(total + .5)
      p(:)  = fo(:) / total
      
c     === cumulative probabilities ===

      xsum = 0.
      do i = 1, ndim
         xsum = xsum + p(i)
         c(i) = xsum
      end do

c     === print observed table ===

      write (2, *) '*** Observed Table ***'
      write (2, *) ''

      call tprint (fo, m, ndim)

c     === criterion value for the observed data ===

      call fcrit  (fo, ndim, m, crit)

      refcrit = crit

      write (2, *) ''
      write (2, 2100) crit
 2100 format('       crit = ' , 3x, f10.4)
      write (2, *) ''
      write (2, *) ''


c     ------------------------------------
c     MAIN LOOP - simulate data nrep times
c     ------------------------------------

      isum = 0

      do j = 1, nrep

c        === generate a random table ===

         call rmultinom (c, x, ndim, n, iseed)
      
c        === evaluate criterion function for this table ===

         call fcrit  (x, ndim, m, crit)

c        === as a check, print the first five simulated tables ===
     
         icount(:) = int(x(:)+.5)
         if (j .le. 5) then
            write (2, 2200, advance='no') icount, sum(icount)
            write (2, 2210) crit
         end if

         if (crit .ge. refcrit) isum = isum + 1

2200  format(20i4)
2210  format(f10.4)
2220  format(f12.6)

c        ===notify user each time another 2500 replications completed===

         if (mod(j, 2500) == 0) then
            write (*, 6000) j
 6000       format(i6)
         end if

         vec(j) = crit

         if (rprint) then
            write (3, 2220) vec(j)
         end if

      end do

c     ----------------
c     end of MAIN LOOP
c     ----------------


c     === sort vector of results ===

      vec2(:) = vec(:)
 
      call shell (nrep, vec2)   

c     ===  write sorted vector ===

      do i = 1, nrep
         write (4, 2220) vec2(i)
      end do

      write (2, *) ''
      write (2, *) '*** Simulation Results ***'
      write (2, *) ''

      write (2, 2310) iseed0
      write (2, 2320) nrep
      write (2, 2330) isum
      write (2, 2340) float(isum)/float(nrep)
      write (2, *) ' '


c     ---------------------     
c     WRITE SUMMARY RESULTS
c     ---------------------

c     === statistics of empirical distribution of crit ===

      cmean  = sum(vec2)/float(nrep)
      cmed   = (vec2(50000)+vec2(50001))/2.

      css    = sum((vec2 - cmean)**2)
      cvar   = css/float(nrep)
      cse    = cvar ** .5

      cl_90l = vec2( 5000)
      cl_90u = vec2(95000)

      cl_95l = vec2( 2500)
      cl_95u = vec2(97500)

      write (2, 2110) refcrit
      write (2, 2120) cmean
      write (2, 2130) cmed
      write (2, 2140) cse
      write (2, 2410) cl_90l, cl_90u
      write (2, 2420) cl_95l, cl_95u

 2110 format('Criterion value            :', f10.4)
 2120 format('Bootstrap mean             :', f10.4)
 2130 format('Bootstrap median           :', f10.4)
 2140 format('Bootstrap standard error   :', f10.4)
 2410 format('90% BS confidence limits   :   (', f6.4, ' - ',f6.4,')' )
 2420 format('95% BS confidence limits   :   (', f6.4, ' - ',f6.4,')' )

c--------------------------------------------------
c  DAG_Stat formula for SE(p+) and SE(p-)
c =SQRT(4*TP*(FP+FN)*(TP+FP+FN))/(TestPos+DisPos)^2
c =SQRT(4*TN*(FP+FN)*(TN+FP+FN))/(TestNeg+DisNeg)^2
c--------------------------------------------------

c     === asymptotic standard error and confidence limits ===

c     == get frequencies as table ==

      indx = 0
      do i = 1, m
      do j = 1, m
         indx = indx + 1
         o(i,j) = fo(indx)
      end do
      end do

      aa = o(1, 1)
      bb = o(1, 2)
      cc = o(2, 1)
      dd = o(2, 2)

      tp = aa 
      tn = dd 
      fn = bb 
      fp = cc 

      testpos = tp + fp
      testneg = tn + fn
      dispos  = tp + fn
      disneg  = fp + tn

      xnum = (4.*TP*(FP+FN)*(TP+FP+FN))**.5
      xden = (TestPos+DisPos)**2
      se2  = xnum / xden

      write (2, *) ''
      write (2, 2150) se2
 2150 format('Asymptotic standard error  :', f10.4)

c     from probit function

      z90 = 1.6452
      z95 = 1.9604

c     from Excel
c     z90 = 1.644853
c     z95 = 1.959961

      acl_90l = refcrit - z90 * se2
      acl_90u = refcrit + z90 * se2

      acl_95l = refcrit - z95 * se2
      acl_95u = refcrit + z95 * se2

      write (2, 2430) acl_90l, acl_90u
      write (2, 2440) acl_95l, acl_95u

 2310 format('random generator seed value:', i10)
 2320 format('Number of simulated tables :', i10)
 2330 format('Number >= reference value  :', i10)
 2340 format('Simulated exact p-value    :', f10.5)
 2430 format('90% AS confidence limits   :   (', f6.4, ' - ',f6.4,')' )
 2440 format('95% AS confidence limits   :   (', f6.4, ' - ',f6.4,')' )

      pause 'Press any key ... '



      end program as



      subroutine fcrit (f, ndim, m, crit)
c------------------------------------------------------
c calculates an arbitrary criterion function of a table
c
c here the criterion is ps+
c------------------------------------------------------

      real      ::      f    (ndim)
      real      ::      o    (m, m)
      real      ::      row  (   m)
      real      ::      col  (   m)
      real      ::      diff (   m)

c     === get frequencies as table ===

      indx = 0
      do i = 1, m
      do j = 1, m
         indx = indx + 1
         o(i,j) = f(indx)
      end do
      end do

c     === row/col marginals ===

      do i = 1, m
         row(i) = sum(o(i,:))
         col(i) = sum(o(:,i))
      end do

      diff(:) = (row(:) - col(:))**2
      rmsd    = (sum(diff)/float(m))**.5

      aa = o(1, 1)
      bb = o(1, 2)
      cc = o(2, 1)
      dd = o(2, 2)

      crit = 2. * aa / (2.* aa + bb + cc)

c     write (3, 3100) aa, bb, cc, dd, crit
 3100 format(4f8.0, f14.6)

      return
      end subroutine fcrit



      SUBROUTINE shell (n,a)

c-----------------------------------------------------------------------------------------------
c  Numerical Recipes in fortran 77, p. 323
c  Sorts an array a(1:n) into ascending numerical order by Shells method (diminishing increment
c  sort). n is input; a is replaced on output by its sorted rearrangement.
c-----------------------------------------------------------------------------------------------
      INTEGER n
      REAL a(n)
      INTEGER i,j,inc
      REAL v

c     === Determine the starting increment ===

      inc = 1 
 1    inc = 3 * inc + 1
      if(inc.le.n) goto 1
 2    continue 

c     === Loop over the partial sorts ===

      inc=inc/3

c     === Outer loop of straight insertion ===

      do 11 i = inc+1, n 
         v = a(i)
         j = i

c        == Inner loop of straight insertion ==

 3       if(a(j-inc).gt.v) then 
            a(j)=a(j-inc)
            j=j-inc
            if(j.le.inc) goto 4
            goto 3
         endif
 4       a(j)=v

11    continue

      if(inc.gt.1) goto 2
      return

      END subroutine shell


      subroutine rmultinom (c, f, ndim, n, iseed)
c--------------------------------------------------------------------------
c rmultinom
c
c produce a random multinomial, f1 ... fm, such that sum f = n,
c according to probabilities, p1 ... p(m)
c
c this version makes n calls to a uniform random number function.
c for large n this is inefficient and another method (such as based
c on generation of random multinomial frequencies) might be
c considered.
c
c author:    john uebersax
c written:   9 july 2006
c
c version:   change(s):
c --------   --------------------------------------------------------------
c
c--------------------------------------------------------------------------
c notes:  when many random multinomials are to be generated given the
c         same {p}, it might be better to supply {c} as input to this
c         subroutine.
c--------------------------------------------------------------------------
      integer          ::     ndim      !  dimension of the multinomial
      integer          ::     n         !  number of observations
      
      real             ::     f (ndim)  !  the random multinomial
      real             ::     c (ndim)  !  cumulative probabilities
      
c     -----------------
c     initialize things
c     -----------------
      
      f(:) = 0.

c     ----------------
c     main loop over n
c     ----------------

      do i = 1, n

c        === get a random number ===

         r = ran0(iseed)

c        === what is the corresponding table cell? ===

         do j = 1, ndim
            if (r .lt. c(j)) goto 20
         end do
  20     continue

c     === increment frequency of that cell ===

         f(j) = f(j) + 1.

      end do
      
      return
      end subroutine rmultinom
      

      subroutine tprint (f, m, ndim)
c----------------------------------------------------------------------  
c  print table frequencies, including marginals
c----------------------------------------------------------------------

      real                :: f(ndim)        ! data as vector
      real                :: o(m,m)         ! data as table

      real                :: r (m)          ! row marginals
      real                :: c (m)          ! column marginals

      character (len=10)  ::  dash (10)     ! for output
      character (len=15)  ::  dash2         ! for output

      dash (  1   ) = ' ---------'
      dash (  2: m) = '----------'
      dash (m+1:10) = '    '
      dash2         = '--+------------'

c     === get frequencies as table ===

      indx = 0
      do i = 1, m
      do j = 1, m
         indx = indx + 1
         o(i,j) = f(indx)
      end do
      end do

c     === row/col marginals ===

      do i = 1, m
         r(i) = sum(o(i,:))
         c(i) = sum(o(:,i))
      end do

c     === write results to output ===

      do i = 1, m
         write (2, 2010, advance='no') o(i,:)
         write (2, 2012, advance='no') 
         write (2, 2010) r(i)
      end do

      write (2, 2020, advance='no') dash(1:m)
      write (2, 2022) dash2

      write (2, 2010, advance='no') c
      write (2, 2012, advance='no')
      write (2, 2010) sum(c)

      write (2, *) ''
      write (2, *) '  Marginals'
      write (2, *) ''
      write (2, *) '         Row        Column'
      write (2, *) '  ----------    ----------'
      do i = 1, m
         write (2, 2200) r(i), c(i)
      end do

 2010 format(10f10.3)
 2012 format('  |  ')
 2020 format(10a10)
 2022 format(a15)
 2200 format(1x, f12.4, 2x, f12.4)

      return
      end subroutine tprint 


      function ran0 (idum)
c-------------------------------------------------------------------------
c from Press et al., Numerical Recipes for fortran
c Chapter 7. Random Numbers, p. 270
c
c Minimal random number generator of Park and Miller.  Returns a uniform
c random deviate between 0. and 1.  Set or reset idum to any integer
c value (except the unlikely value mask) to initialize the sequence;
c idum must not be altered between calls for successive deviates
c in a sequence.
c-------------------------------------------------------------------------

      integer ::  idum, ia, im, iq, ir, mask
      real    ::  ran0, am

      parameter (ia = 16807, im = 2147483647, am = 1./im,
     1           iq = 127773,ir = 2836, mask = 123459876 )

      integer k

      idum = ieor(idum, mask)       ! xoring with mask allows use of zero
                                    ! and other simple bit patterns for idum

      k    = idum / iq
      idum = ia*(idum-k*iq)-ir*k    ! compute idum=mod(ia*idum,im) without
                                    ! overflows by Schrage's method

      if (idum .lt. 0) idum=idum+im
      ran0 = am*idum                ! convert idum to a floating result

      idum = ieor(idum,mask)        ! unmask before return
      return
      end function ran0
